import { Injectable } from '@angular/core';
import { HttpClientModule, HttpClient, HttpHeaders, HttpErrorResponse } from '@angular/common/http';
import { Observable, Subject, BehaviorSubject } from 'rxjs';
import { catchError, map, tap, take,filter,skip } from 'rxjs/operators';
import { ActivatedRoute, Router } from '@angular/router';
import {LocationStrategy, Location} from '@angular/common';
// import { keypair } from 'keypair';
import * as keypair from 'keypair';
import * as forge from "node-forge";
import { Identity, AuthToken, KeyCert, SshAuthzServer } from './identity';
import {BackendSelectionService} from './backend-selection.service';
import { throwError } from 'rxjs';
import {NotificationsService} from './notifications.service';


export class SshauthzServer {}

@Injectable({
  providedIn: 'root'
})
export class AuthorisationService {
  // public token: BehaviorSubject<AuthToken>;
  public token: BehaviorSubject<AuthToken>;
  public readonly sshAuthzServers: BehaviorSubject<SshAuthzServer[]>;
  public readonly agentContents: BehaviorSubject<any>;
  public loggedInAuthZ: BehaviorSubject<SshAuthzServer[]>;
  public loggedOutAuthZ: BehaviorSubject<SshAuthzServer[]>;
  public readyToNavigate: Subject<[Boolean,string]>;
  public progress: Subject<string>;
  // private keyCert: Subject<KeyCert>;
  public backendURI: string;
  public keys: KeyCert[];


  constructor(private http: HttpClient,
              private locationStrategy: LocationStrategy,
              private route: ActivatedRoute,
              private router: Router,
              private backendSelectionService: BackendSelectionService,
              private location: Location, 
              private notifications: NotificationsService) {
                this.token = new BehaviorSubject<AuthToken>(null);
                this.readyToNavigate = new Subject<[Boolean,string]>();
                this.readyToNavigate.next([false,'']);
                this.progress = new Subject<string>();
                this.progress.next("");
                // this.token.subscribe(token => this.getCert(token));
                this.route.fragment.subscribe(frag => this.storeToken(frag));
                this.sshAuthzServers = new BehaviorSubject<SshAuthzServer[]>([]);
                this.loggedInAuthZ = new BehaviorSubject<SshAuthzServer[]>(null);
                this.loggedOutAuthZ = new BehaviorSubject<SshAuthzServer[]>([]);
                this.backendURI = null;
                this.agentContents = new BehaviorSubject(null);
                this.keys = [];
                this.getSshAuthzServers();
                this.keys = [];
                
    this.backendSelectionService.apiserver.pipe(filter((v) => v !== null && v !== undefined))
      .subscribe((value) => { this.backendURI = value.tes ; this.agentContents.next(null) })
    this.agentContents.subscribe((value) => this.updateLoggedAuthZ());
    this.sshAuthzServers.subscribe((value) => this.updateLoggedAuthZ());
 }


 storeLocalAuthZ(authz: any) {
   try {
     localStorage.setItem('localauthservers',JSON.stringify(authz));
   } catch {
   }
   this.getSshAuthzServers();
 }
 removeLocalAuthZ() {
   localStorage.removeItem('localauthservers');
   this.getSshAuthzServers();
 }

 getSshAuthzServers() {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: false};
   this.http.get<SshAuthzServer[]>('./assets/config/authservers.json',options)
                    .pipe(catchError(this.handleError('getSshAuthzServers')))
                    .subscribe(resp => this.updateSshAuthzServers(resp));
 }

 private updateLoggedAuthZ() {
     let agentContents = this.agentContents.value
     let authzServers = this.sshAuthzServers.value
     let loggedin = []
     let loggedout = []
     var found: boolean;
     if (agentContents == null) {
       return
     }
     for (let s of authzServers) {
         found=false;
         for (let cert of agentContents) {
           if ('Signing CA' in cert) {
             for (let ca of cert['Signing CA']) {
               if (ca == s.cafp) {
                 loggedin.push(s)
                 found=true;
                 continue;
               }
             }
           }
           if (found) {
               continue;
           }
         }
         if (!found)  {
           loggedout.push(s)
         }
     }
     this.loggedOutAuthZ.next(loggedout);
     this.loggedInAuthZ.next(loggedin);
 }
        

 updateSshAuthzServers(resp) {
   var auths: SshAuthzServer[];
   var localauths: SshAuthzServer[] = [];
   var server: SshAuthzServer;
   auths = <SshAuthzServer[]>resp;
   try {
       localauths = JSON.parse(localStorage.getItem('localcomputesites'))
   } catch {
       localauths = []
   }
   if (localauths !== null) {
       for (server of localauths) {
           auths.push(server);
       }
   }
   this.sshAuthzServers.next(auths);
 }

 extractToken(frag: string) {
   if (frag === undefined || frag == null) {
       return;
   }
   let tokenmatch = null;
   let statematch = null;
   if (!(frag === undefined) && !(frag == null)) {
     tokenmatch = frag.match(/access_token\=([^&]+)(&|$)/);
     statematch = frag.match(/state\=([^&]+)(&|$)/);
   }
   if (tokenmatch == null || statematch == null) {
     throw new Error('no token present');
   }

   let accesstoken = tokenmatch[1];
   let state = statematch[1];
   // this.router.navigate(['/']);

   //Verify that the state matched the nonce we used when initiating login
   let tuple = JSON.parse(sessionStorage.getItem('authservice'));
   if (tuple[1] != state) {
     throw new Error('callback state parameter does not match');
   }

   //throw new Error('test error handling');
   return new AuthToken(tokenmatch[1],tuple[0]);

 }

 storeToken(frag: string) {
   if (frag === undefined || frag == null) {
       return;
   }
   let tokenmatch = null;
   let statematch = null;
   if (!(frag === undefined) && !(frag == null)) {
     tokenmatch = frag.match(/access_token\=([^&]+)(&|$)/);
     statematch = frag.match(/state\=([^&]+)(&|$)/);
   }
   if (tokenmatch == null || statematch == null) {
     return;
   }

   let accesstoken = tokenmatch[1];
   let state = statematch[1];
   // this.router.navigate(['/']);

   //Verify that the state matched the nonce we used when initiating login
   let tuple = JSON.parse(sessionStorage.getItem('authservice'));
   if (tuple[1] != state) {
     return
   }

   this.token.next(new AuthToken(tokenmatch[1],tuple[0]));
   //this.token = new AuthToken(tokenmatch[1],tuple[0]);


 }

 public getCert(token: AuthToken, apiserver: any, pubkey: any) {
   /*var token: AuthToken;
   token = this.token.value;
   if (token === null || token === undefined || token.token === undefined || token.token === '' || token.token === null) {
     this.token.pipe(skip(1),take(1)).subscribe(() => this.getCert());
     return
   }
   if (this.backendSelectionService.apiserver.value === null || this.backendSelectionService.apiserver.value === undefined) {
     this.backendSelectionService.apiserver.pipe(skip(1),take(1)).subscribe(() => this.getCert());
     return
   }*/


   /*let starttime = new Date();
   let newkeypair = keypair();
   let publicKey = forge.pki.publicKeyFromPem(newkeypair.public);
   let sshpub = forge.ssh.publicKeyToOpenSSH(publicKey, 'sv2@monash.edu');
   let endtime = new Date();*/

   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};

   let data = {'token': token.token, 'pubkey': pubkey, 'signing_url': token.sshauthzservice.sign};

   return this.http.post<any>(apiserver.tes+'/getcert',data, options)
   //.pipe(catchError(this.handleError([])))
   //                   .subscribe(resp => this.makeKeyCert(newkeypair.private, resp, token.sshauthzservice),
   //                              error => this.signingError(error,token.sshauthzservice));
 }

public getKeys(id?: Identity) {
  try{
    return JSON.parse(sessionStorage.getItem('keys'));
  } catch {
    return [];
  }
}

 makeKeyCert(key: string, resp, sshauthzservice: SshAuthzServer) {
   let keyCert = new KeyCert()
   keyCert.key = key;
   keyCert.cert = resp['cert'];
   var keys: KeyCert[] = [];
   try{
     keys = JSON.parse(sessionStorage.getItem('keys'));
   } catch {
     keys = [];
   }
   if (keys === null) {
     keys = [];
   }
   keys.push(keyCert);
   sessionStorage.setItem('keys',JSON.stringify(keys))
   // this.tesService.keyCert.next(keyCert);
   // As soon as the certificate has been generated, we log back out of the signing server
   if (!(sshauthzservice.logout === null)) {
     window.open(sshauthzservice.logout);
   }
   let path=sessionStorage.getItem('path');
   //skip1 because loggedInAuthZ is a behaviour subject and we don't want the current value but the value
   this.loggedInAuthZ.pipe(skip(1),take(1)).subscribe( () => {this.readyToNavigate.next([true,path])});
   this.sshAdd(keyCert);
   // only navigate once the agent contents has been refreshed
 }

 public querySshAgentError(error: any) {
   this.agentContents.next([]);
   if (error.status == 0) {
     this.notifications.notify("A network error occured. Are you connected to the internet?")
   }
   this.notifications.notify("Error querying ssh agent");
 }

 public updateAgentContents() {
   if (this.backendURI == null) {
     return
   }
   this.notifications.notify("Updating the list of available accounts");
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   var anyvar: any;
   this.http.get<any>(this.backendURI+'/sshagent',options)
                     .pipe(catchError(this.handleError(anyvar)))
                     .subscribe(resp => { this.agentContents.next(resp); this.notifications.notify("") },
                                error => this.querySshAgentError(error));
                     // .subscribe(resp => this.computeSitesService.updateIdentities(resp),
                     //            error => this.httperror(error))
 }


 private killAgent() {
   this.notifications.notify("Logging out")
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   var anyvar: any;
   this.http.delete<any>(this.backendURI+'/sshagent',options)
                     .pipe(catchError(this.handleError(anyvar)))
                     .subscribe(resp => this.updateAgentContents(),
                                error => this.httperror(error));
 }

 public logout(): Boolean {
   this.killAgent();
   return true;
 }



 public sshAdd(keyCert: KeyCert) {
   if (keyCert.key == undefined) {
     return;
   }
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   var anyvar: any;

   this.notifications.notify("Authorising ...")
   let data = {'key': keyCert.key, 'cert': keyCert.cert};
   this.http.post<any>(this.backendURI+'/sshagent',data,options)
                       .pipe(catchError(this.handleError(anyvar)))
                       .subscribe(resp => this.updateAgentContents(),
                                  error => this.httperror(error))
 }

 public login(authservice: SshAuthzServer) {
   let redirect_uri = window.location.origin+this.locationStrategy.getBaseHref()+"sshauthz_callback";
   let nonce=Math.random().toString(36).substring(2, 15)

   sessionStorage.setItem('authservice', JSON.stringify([authservice,nonce]));
   sessionStorage.setItem('path', '/launch');
   if (authservice.scope == null) {
     window.location.assign(authservice.authorise+"?response_type=token&redirect_uri="+redirect_uri+"&state="+nonce+"&client_id="+authservice.client_id);
   } else {
     window.location.assign(authservice.authorise+"?response_type=token&redirect_uri="+redirect_uri+"&state="+nonce+"&client_id="+authservice.client_id+"&scope="+authservice.scope);
   }
 }

 private httperror(error: any) {
   this.notifications.notify('There was an error logging in or generating crypto tokens');
   console.error(error);
 }
 private signingError(error: any,sshauthzservice: SshAuthzServer) {
   this.notifications.notify('You don\'t appear to have an account on '+sshauthzservice.name);
   if (!(sshauthzservice.logout === null)) {
     window.open(sshauthzservice.logout);
   }
   this.readyToNavigate.next([true,'/noaccount/'+sshauthzservice.name])
 }
 

 private handleError<T> (result?: T) {
   return (error: any): Observable<T> => {
     if (error.status == 500) {
       return throwError("The backend server encountered and error. Please try again in a few minutes")
     }
     return throwError(error.message);
     // return of(result as T);
   };
 }
}
