import { Component, OnInit, NgModule, Inject, Input } from '@angular/core';
import { MatTableDataSource, MatDialog, MatDialogRef, MAT_DIALOG_DATA, MatDialogModule, MatDialogConfig } from '@angular/material';

import { Location } from '@angular/common';
// import { HttpClientModule, HttpClient, HttpHeaders, HttpErrorResponse } from '@angular/common/http';
import { Observable, Subject, pipe } from 'rxjs';
import { Subscription, of } from 'rxjs';
import { fromEvent } from 'rxjs';
import { timer, combineLatest } from 'rxjs';
import { repeat, takeUntil, skip, switchMap, take, filter} from 'rxjs/operators';
import { Router, ActivatedRoute, ParamMap } from '@angular/router';

import {Strudelapp} from '../strudelapp';
import { StrudelappsService } from '../strudelapps.service';
import { TesService } from '../tes.service';
import {BackendSelectionService } from '../backend-selection.service';
import { AuthorisationService } from '../authorisation.service';
import { Identity } from '../identity';
import { Computesite, Health } from '../computesite';
import { LogoutdialogComponent } from '../logoutdialog/logoutdialog.component';

import { SshAuthzServer } from '../identity';
import { ComputesitesService } from '../computesites.service';
import { BehaviorSubject } from 'rxjs/BehaviorSubject';
import { SettingsService} from '../settings.service';




export interface Quota {
  resource: string;
  usage: string;
  quota: string;
}




@Component({
  selector: 'app-launcher',
  templateUrl: './launcher.component.html',
  styleUrls: ['./launcher.component.css']
})
export class LauncherComponent implements OnInit {
  public strudelapps: Strudelapp[];

  public app$: BehaviorSubject<Strudelapp>;
  public authorised: boolean;
  public identity: Identity;
  public identity$: BehaviorSubject<Identity>;

  public identities: Identity[];
  public sshauthzservers: SshAuthzServer[];
  private launchwindow: any;
  private launchwindowWatcher: any;
  private subscriptions: Subscription[];
  //public quotas: BehaviorSubject<any[]>;
  displayedColumns: string[] = [];
  private app: Strudelapp;


  constructor( public dialog: MatDialog,
                public tesService: TesService,
                public backendSelectionService: BackendSelectionService,
                public authService: AuthorisationService,
                public computeSitesService: ComputesitesService,
                private router: Router,
                private route: ActivatedRoute,
                private strudelappsService: StrudelappsService,
                private settingsService: SettingsService,
                ) {
      this.subscriptions = [];
      this.subscriptions.push(this.authService.sshAuthzServers.subscribe(o => {this.updateSshAuthZServers(o)}));
      this.subscriptions.push(this.computeSitesService.identities.subscribe(o => this.navLogin(o)));
      this.subscriptions.push(this.computeSitesService.appidentities.subscribe(o => this.getHealth(o)));
      this.app$ = new BehaviorSubject<Strudelapp>(null);
      this.identity$ = new BehaviorSubject<Identity>(null);
      this.identity = null;
    //this.quotas = new BehaviorSubject<any[]>([]);
  }

  countErrors(a: Health[], b: Health[]) {
    var count: number = 0
    var h: Health;
    if (a != null) {
      for (h of a) {
        if (h.stat == 'error' || h.stat == 'warn') {
          count++;
        }
        if(!isNaN(parseInt(h.stat))) {
          count = count + parseInt(h.stat);
        }
      }
    }
    if (b != null) {
      for (h of b) {
        if (h.stat == 'error' || h.stat == 'warn') {
          count++;
        }
        if(!isNaN(parseInt(h.stat))) {
          count = count + parseInt(h.stat);
        }
      }
    }
    return count;
  }

  calculateCols(data) {
    return data.map(function(r) { return r.key });
  }

  humanKBytes(n: number) {
    if (n > 1024*1024*1024) {
      let v = n/1024/1024/1024;
      return v.toFixed(0)+'TB';
    }
    if (n > 1024*1024) {
      let v = n/1024/1024;
      return v.toFixed(0)+'GB';
    }
    if (n > 1024) {
      let v = n/1024;
      return v.toFixed(0)+'MB';
    }
    return '0 MB';
  }

  rowClass(row) {
    if (row.stat == 'error' || row.stat == 'warn') {
      return 'health-warn';
    } else {
      return 'health-ok';
    }
  }

  navLogin(o) {
    if (o == null) {
      return
    }
    if (o.length == 0) {
      console.log('in navLogin');
      this.router.navigate(['/login']);
    } else {
    }
  }

  getHealth(o) {
    if (o == null) {
      return
    }
    this.tesService.cancelHealthRequests();
    for (let id of o) {
      this.tesService.getHealthAlerts(id);
    }
  }

  updateSshAuthZServers(o) {
    this.sshauthzservers = o;
  }

    

  ngOnInit() {
    const notNull = pipe(filter((v) => v !== null));
    this.strudelapps = [];
    combineLatest([this.route.paramMap,this.computeSitesService.appidentities.pipe(notNull)]).subscribe(([map,ids]) => this.updateIdApp(<ParamMap> map,<Identity[]> ids));
    //this.route.paramMap.subscribe((params: ParamMap) => this.updateIdApp(params));
    // if the conents of the ssh agent changes we will get a whole new set of identity objects. In that case we must update out id subject
    this.computeSitesService.appidentities.pipe(notNull).subscribe( (appidentities) => this.refreshId(<Identity[]> appidentities));
  }


  refreshId(appidentities: Identity[]) {
    var ids: string;
    var id: Identity;
    if (this.identity$.value !== null) {
      ids = this.identity$.value.site.name;
    } else {
      return
    }
    id = this.getId(ids,appidentities);
    this.identity$.next(id);
    if (id === null) {
      if (appidentities.length == 0) {
        this.identity$.next(null);
        console.log('in refreshid');
        this.router.navigate(['/login']);
        return;
      } else {
        this.router.navigate(['/launch',this.computeSitesService.appidentities.value[0].displayName(),'accountinfo']);
        return
      }
    }
  }

  updateIdApp(params: ParamMap, appidentities: Identity[]) {
    var idsite: string = params.get('site');
    var apps: string = params.get('app');
    if (appidentities === null) {
      return 
    }
    if (idsite === null) {
      if (appidentities.length == 0) {
        this.identity$.next(null);
        this.router.navigate(['/login']);
        return;
      } else {
        this.router.navigate(['/launch',this.computeSitesService.appidentities.value[0].displayName(),'accountinfo']);
        return
      }
    }
    var id: Identity = this.getId(idsite,appidentities);
    if (id === null) {
      if ( appidentities.length == 0) {
        this.identity$.next(null);
        console.log('in updateIdApp 2');
        this.router.navigate(['/login']);
        return;
      } else {
        this.router.navigate(['/launch',this.computeSitesService.appidentities.value[0].displayName(),'accountinfo']);
        return
      }
      //this.computeSitesService.appidentities.pipe(skip(1),filter((v) => v !== null && v.length > 0),take(1)).subscribe(() => this.updateIdApp(params));
      return
    }
    if (id !== this.identity$.value) {
      this.identity$.next(id);
    }     

    if (apps === null) {
      apps = 'accountinfo';
    }
    if (id.site.appCatalog.value === undefined || id.site.appCatalog.value === null || (id.site.appCatalog.value).length == 0) {
      id.site.appCatalog.pipe(filter((v) => v !== null && v.length > 0),take(1)).subscribe(() => this.updateIdApp(params,appidentities));
      return
    }
    var app: Strudelapp = this.getApp(id,apps);
    if (app !== this.app$.value) {
      this.app$.next(app);
    }
  }

  getApp(id,v: any): Strudelapp {
    if (id == 'accountinfo') {
      return null;
    }
    return this.strudelappsService.getApp(v,id.site.appCatalog.value);
  }

  getId(v: any, appidentities: Identity[]) {
    if (v === null) {
      console.error('getId called with null value');
      return null;
    }
    if (appidentities == []) {
      console.error('no appidentities yet? call back latter');
      return null
    }

    for ( let id of appidentities) {
      if (v == id.displayName()) {
        return id
      } 
    }

    for ( let id of appidentities) {
      if (v == id.site.name) {
        return id
      }
    }

    var idx: number;
    idx = v.indexOf('@');
    var sitename: string;
    if (idx > 0) {
      sitename = v.substring(idx+1);
    } else {
      sitename = v;
    }

    for ( let id of this.computeSitesService.appidentities.value) {
      if (sitename == id.site.name) {
        return id
      } else {
        console.log('sitename',sitename,'is not the same as ',id.site.name);
      }
    }

    console.error('didnt find a matching id');
    console.log(v);
    console.log(this.computeSitesService.appidentities.value);
    return null;
  }


  ngOnDestroy() {
    for (let s of this.subscriptions) {
      s.unsubscribe();
    }
  }

  selectId(id: Identity) {
    this.identity$.next(id);
    this.identity = id;
    this.app$.next(null);
  }

  selectApp(app: Strudelapp) {
    this.app$.next(app);
  }


}
