import { Component, OnInit, AfterViewInit } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import {  map, switchMap, filter, catchError, tap } from 'rxjs/operators'
import { combineLatest, of, pipe, timer } from 'rxjs';
import {BackendSelectionService} from '../backend-selection.service';

import * as keypair from 'keypair';
import * as forge from "node-forge";

import { AuthorisationService } from '../authorisation.service';




@Component({
  selector: 'app-keygen',
  templateUrl: './keygen.component.html',
  styleUrls: ['./keygen.component.css']
})
export class KeygenComponent implements OnInit {

  constructor(private route: ActivatedRoute, private authService: AuthorisationService, private backendSelectionService: BackendSelectionService) { }

  ngOnInit() {
    timer(100).subscribe( _ => this.initPipelines());
  }
 
  initPipelines() {
    console.log('in init pipelines');
    let data$ = this.route.fragment.pipe( // When the token is available, generate a new key
      tap( _ => console.log('begininig pipeline')),
      map((v) => this.authService.extractToken(v)),
      switchMap((token) => of(keypair()),
        (token,kp,iter1,iter2) =>  { return {'token': token, 'key': kp, 'iter1': iter1, 'iter2': iter2}} // because keypair is slow we will create another observable
      ),
      tap( _ => console.log('switchMap\'d')),
      map((data) => {data['pubkey'] = this.getpubkey(data['key']) ; return data }),
    );

    let apiserver$ = this.backendSelectionService.apiserver.pipe(filter((v) => v !== undefined),filter((v) => v !== null));

    //data$.subscribe((data) => {console.log('generated kp'); console.log(data)},
    //                 (err) => { console.log('caught in subscription',err)} );
    // Once the key is available, and there is a backend server to use,
    // generate a cert, and add it to the ssh agent
    let combo = combineLatest([data$,apiserver$])
      .pipe(
        map(([data,apiserver]) => { data['apiserver'] = apiserver; return data }),
        tap((v) => console.log('tap1',v)),
        switchMap(
          (data) => this.authService.getCert(data['token'],data['apiserver'],data['key']['pubkey']),
          (data,certresponse,idx1,idx2) => { data['cert'] = certresponse ; return data }),
        tap((v) => console.log('tap2',v))
      );
    combo.subscribe((v) => console.log('auth complete'));
  }


  getpubkey(kp) {
    return forge.ssh.publicKeyToOpenSSH(
      forge.pki.publicKeyFromPem(kp.public), 'sv2@monash.edu')
  };

    /*  ngAfterViewInit() {
    this.authService.getCert();
  }

  genKey() {
   let newkeypair = keypair();
   let publicKey = forge.pki.publicKeyFromPem(newkeypair.public);
   let sshpub = forge.ssh.publicKeyToOpenSSH(publicKey, 'sv2@monash.edu');
   return sshpub
  }*/

    /*  navigate(readyToNavigate: [Boolean,string]) {
    if (readyToNavigate[0]) {
      this.router.navigate([readyToNavigate[1]]);
    }
  }*/

  }
